#include <iostream>
#include <fstream>
#include <iomanip>
#include <list>
#include <string>

#include "binarystreams.h"

using namespace std;

// In this example, a MemHandle is just an offset into the file.
typedef int MemHandle;

// The memory manager interface from the textbook.  One addition not found in
// the textbook:  if a null pointer is passed to the "space" parameter of the
// get() method, no data is actually read -- only the length of the block
// stored at that handle is returned.  This can be used in situations where
// you deal with variable length records:  to read the record, you do the
// following:
//
//   1) Call get() with a NULL pointer to get the length of the block
//   2) Allocate a block of memory equal to that size
//   3) Call get() again, with a pointer to the newly allocated block to read
//      the actual data
class MemManager
{
public:
	virtual MemHandle insert(void* space, int length) = 0;
	virtual void release(MemHandle h) = 0;
	virtual int get(void* space, MemHandle h) = 0;
};

// A very basic memory manager that doesn't implement any actual memory
// management -- it just inserts new data to the end of the file.
class SimpleMemManager : public MemManager
{
private:
	fstream file;
	
public:
	SimpleMemManager(const char* path)
	{
		file.open(path, ios::in | ios::out | ios::binary | ios::trunc);
	}
	
	MemHandle insert(void* space, int length)
	{
		file.seekp(0, ios::end);
		MemHandle handle = file.tellp();
		
		file.write((char*)&length, sizeof(int));
		file.write((char*)space, length);
		
		return handle;
	}
	
	void release(MemHandle h) { }
	
	int get(void* space, MemHandle h)
	{
		int length;

		file.seekg(h);
		file.read((char*)&length, sizeof(int));

		if(space != 0)
			file.read((char*)space, length);
		
		return length;
	}
};

// The Serializable interface defines the methods that will be used to read
// and write objects from the memory manager.
class Serializable
{
public:
	virtual MemHandle writeObject(MemManager* manager) = 0;
	virtual void readObject(MemManager* manager, MemHandle handle) = 0;
};

// A simple class requiring custom serialization because string and list fields
// can in no way be directly written to disk.
//
// By using custom serialization, we can determine completely the layout of the
// record on disk.  In this case, we will write the record as follows:
//
//   * Length of the "name" field + terminating null (int = 4 bytes)
//   * Characters comprising "name" + terminating null (name.length() + 1 bytes)
//   * Number of elements in "phoneNumbers" list (int = 4 bytes)
//   * All phone numbers in "phoneNumbers" list, in sequence
//       (phoneNumbers.size() * 4 bytes)
//
class Contact : public Serializable
{
public:
	string name;
	list<int> phoneNumbers;

	// Write the object to the memory manager.  By using the obinarystream, we
	// don't need to pre-allocate the memory for the object.  Instead, we
	// write the elements to the stream in sequence, which causes the stream's
	// buffer to grow as necessary.  Finally, we send the memory associated
	// with the stream to the memory manager to be written to the disk.
	MemHandle writeObject(MemManager* manager)
	{
		obinarystream os;
		os << name.length() + 1;

		// Note that we have to write the string a character at a time -- if we
		// pass the string object directly to the stream, then it will write
		// the raw binary contents of the string object, including pointers to
		// the character data and other state-specific fields.
		for(int i = 0; i < name.length(); i++)
			os << name[i];

		// Notice that we write the terminating null character explicitly.  We
		// actually don't have to do this, since the readObject() function
		// doesn't make use of it, but I wanted to keep the record layout the
		// same as the example in serialization1.cpp.
		os << '\0';

		// Instead of storing an actual list structure on disk, we just output
		// the elements in sequence.
		os << phoneNumbers.size();
		for(list<int>::iterator it = phoneNumbers.begin(); it != phoneNumbers.end(); ++it)
		{
			int phoneNum = *it;
			os << phoneNum;
		}
		
		// Send the data to the memory manager.  The buffer inside the
		// obinarystream will be automatically freed when the stream goes out
		// of scope.
		MemHandle handle = manager->insert(os.data(), os.size());
		
		return handle;
	}

	// Reading the object from disk undoes the operation above.  First, we
	// request from the memory manager the size of the data at the specified
	// handle.  Then, we create an ibinarystream of the appropriate size, load
	// the object data from the memory manager into the stream's buffer, and
	// pull the data we're interested in off the stream.
	void readObject(MemManager* manager, MemHandle handle)
	{
		int objLength = manager->get(0, handle);
		ibinarystream is(objLength);
		manager->get(is.data(), handle);
		
		int nameLength, phoneCount;
		char ch;
		
		// Pull off the length of the name string.
		is >> nameLength;
		
		// Read the string a character at a time from the stream (recall that
		// the length written to the disk includes the terminating null
		// character, so we subtract one here).
		name.clear();
		for(int i = 0; i < nameLength - 1; i++)
		{
			is >> ch;
			name += ch;
		}
		
		// Swallow the terminating null, which we don't use in this case.
		is >> ch;
		
		// Pull off the number of phone numbers.  We'll read this many
		// integers from the stream and build up the linked list from this
		// sequence.
		is >> phoneCount;
		phoneNumbers.clear();
		for(int i = 0; i < phoneCount; i++)
		{
			int phoneNum;
			is >> phoneNum;
			phoneNumbers.push_back(phoneNum);
		}
	}
};

int main()
{
	SimpleMemManager manager("testMemMan.bin");

	// Create a couple simple Contact objects and write them to the file.
	Contact* contact1 = new Contact();
	contact1->name = "Vincent Schiavelli";
	contact1->phoneNumbers.push_back(5551234);
	contact1->phoneNumbers.push_back(9876543);

	MemHandle contact1Handle = contact1->writeObject(&manager);
	delete contact1;

	Contact* contact2 = new Contact();
	contact2->name = "A Really Long Name That Will Take Up A Considerable Amount of Space";
	contact2->phoneNumbers.push_back(1111111);
	contact2->phoneNumbers.push_back(1111112);
	contact2->phoneNumbers.push_back(1111113);
	contact2->phoneNumbers.push_back(1111114);
	contact2->phoneNumbers.push_back(1111115);
	contact2->phoneNumbers.push_back(1111116);
	contact2->phoneNumbers.push_back(1111117);
	contact2->phoneNumbers.push_back(1111118);
	contact2->phoneNumbers.push_back(1111119);
	contact2->phoneNumbers.push_back(1111110);

	MemHandle contact2Handle = contact2->writeObject(&manager);
	delete contact2;

	// Read the data back from the memory manager and verify that the objects
	// have been properly reconstructed.
	Contact* contact2Read = new Contact();
	contact2Read->readObject(&manager, contact2Handle);
	cout << "Contact 2 name = " << contact2Read->name << endl;
	cout << "Contact 2 phone numbers:" << endl;
	for(list<int>::iterator it = contact2Read->phoneNumbers.begin(); it != contact2Read->phoneNumbers.end(); ++it)
	{
		cout << "   " << *it << endl;
	}
	cout << endl;
	delete contact2Read;

	Contact* contact1Read = new Contact();
	contact1Read->readObject(&manager, contact1Handle);
	cout << "Contact 1 name = " << contact1Read->name << endl;
	cout << "Contact 1 phone numbers:" << endl;
	for(list<int>::iterator it = contact1Read->phoneNumbers.begin(); it != contact1Read->phoneNumbers.end(); ++it)
	{
		cout << "   " << *it << endl;
	}
	cout << endl;
	delete contact1Read;

	return 0;
}
