#include <iostream>
#include <fstream>
#include <iomanip>

using namespace std;

class Rec
{
private:
	// Private constructors and destructors prevent code from creating record
	// objects through normal means (local variables, standard calls to new).
	// Instead, construction and destruction are handled entirely through the
	// static methods below.
	Rec(int id, double gpa, short sal)
	{
		ID = id;
		GPA = gpa;
		salary = sal;
	}

	~Rec() { }

public:
	// This static method allocates a block of memory equal to the size of the
	// fields in the record PLUS the amount of space requested for the variable
	// length array at the end of the structure.  Then the "placement new"
	// operator is used to construct the object in the pre-allocated space.
	static Rec* construct(int id, double gpa, short sal, int stuffLen)
	{
		// sizeof(Rec) takes into account the size of all of the fields below,
		// except for "char stuff[]", whose size is technically zero
		// (undefined).  The parameter "stuffLen" in this case specified how
		// much memory we want to allocate "at the end of the object" to store
		// additional data.
		char* space = new char[sizeof(Rec) + stuffLen];
		
		// Since we have already allocated the amount of space we want for the
		// object, the syntax "new(space) ..." instead of just "new" means that
		// the object will be initialized in the space provided, instead of
		// allocating new memory.  This allows us to pass the other parameters
		// (id, gpa, sal) to the private constructor above to handle other
		// initialization.
		//
		// The pointer returned by this statement actually points to the very
		// same memory pointed to by "space".
		Rec* rec = new(space) Rec(id, gpa, sal);
		return rec;
	}

	// Using "placement new" means that we have to take special care when
	// destroying the object.  We cannot simply say "delete rec", because the
	// memory was allocated as an array.  Instead, we call the destructor
	// explicitly, then cast the pointer back to a char* and delete it in a way
	// symmetric to that in which it was allocated.
	void destroy()
	{
		this->~Rec();
		delete[] reinterpret_cast<char*>(this);
	}

	int ID;
	double GPA;
	int salary;
	char major[4];
	char stuff[];	// An array of undeclared length can only appear at the
					// end of a structure/class, and there can only be one such
					// field.
};

void hexdump(char* ptr, int size);
  
int main()
{
	const int STUFF_LENGTH = 36;	// the amount of extra space we want to
									// reserve for the stuff[] field
	
	int id = 111223333;
	double gpa = 2.345;
	int salary = 2222;
	char major[4] = { 'C', 'S', 'E', 'D' };
	char stuff[STUFF_LENGTH] = { 0 };
	strncpy(stuff, "John Doe:1002 Anywhere Street", STUFF_LENGTH);


	ofstream myFile("data4.bin", ios::out | ios::binary);

	// The constructor of Rec was made private to prevent direct allocation
	// via "new Rec(...);"  Instead, we force the user to call our special
	// allocation function, which requires that we pass the length of the extra
	// space we need for the stuff[] field.
	Rec* myrecord = Rec::construct(id, gpa, salary, STUFF_LENGTH);

	// strcpy(myrecord->major,major); <-- This would be dangerous!!!  major[]
	//                                    is NOT null-terminated, so we cannot
	//                                    use strcpy

	memcpy(&(myrecord->major), major, 4 * sizeof(char));
	memcpy(&(myrecord->stuff), stuff, STUFF_LENGTH * sizeof(char));

	// Verify the contents of all the fields in the object.
	cout << "myrecord->ID     = " << myrecord->ID << endl;
	cout << "myrecord->GPA    = " << myrecord->GPA << endl;
	cout << "myrecord->salary = " << myrecord->salary << endl;
	cout << "myrecord->major  = ";
	for(int i = 0; i < 4; i++)
		cout << myrecord->major[i];
	cout << endl;

	// Notice here that we can access myrecord->stuff as if it were any other
	// char array.
	cout << "myrecord->stuff  = " << myrecord->stuff << endl;
	cout << endl;

	// Now write the entire record to the file at once.
	myFile.write((char*)myrecord, sizeof(Rec) + STUFF_LENGTH);
	myFile.close();


	// Now we'll read the data back from the file, in separate operations.
	ifstream myRead("data4.bin",ios::in | ios::binary);
	Rec* readRecord = Rec::construct(0, 0, 0, STUFF_LENGTH);
	myRead.read((char*)readRecord, sizeof(Rec) + STUFF_LENGTH);

	// Verify that the data read from disk is correct.
	cout << "Printing the data read from disk:" << endl;
	cout << "id     = " << readRecord->ID << endl;
	cout << "gpa    = " << readRecord->GPA << endl;
	cout << "salary = " << readRecord->salary << endl;
	cout << "major  = ";
	for(int i = 0; i < 4; i++)
		cout << readRecord->major[i];
	cout << endl;
  	cout << "stuff  = " << readRecord->stuff << endl;
  
	cout << endl;

	cout << "Dump of raw bytes occupied by the record we created:" << endl;
	hexdump((char*)myrecord, sizeof(Rec) + STUFF_LENGTH);
	cout << endl;

	cout << "Dump of raw bytes occupied by the record we read back:" << endl;
	hexdump((char*)readRecord, sizeof(Rec) + STUFF_LENGTH);

	// We can't say "delete myrecord;" because it would circumvent the custom
	// allocation done in the Rec::construct method (and in fact, we made the
	// destructor private to turn this into a compile-time error).  Instead, we
	// provide a custom "destroy" function that mirrors the functionality of
	// the destructor, but properly frees the object.
	myrecord->destroy();
	readRecord->destroy();

	return 0;
} // end main


// Print a hex-editor-like view of the memory at the specified address.
void hexdump(char* ptr, int size)
{
	cout << hex << setfill('0');

	for(int offset = 0; offset < size; offset += 16)
	{
		cout << setw(4) << offset << ": ";

		int bytesOnRow = min(16, size - offset);

		for(int i = 0; i < 16; i++)
		{
			if(i < bytesOnRow)
				cout << setw(2) << (int)(unsigned char)ptr[offset + i] << ' ';
			else
				cout << "   ";
		}

		cout << ' ';
		for(int i = 0; i < 16; i++)
		{
			if(i < bytesOnRow)
			{
				char ch = ptr[offset + i];
				cout << ((32 <= ch && ch <= 126) ? ch : '.');
			}
		}
				
		cout << endl;
	}
}
